/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.service.flexible.tools;

import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlItemTypeEnum;
import org.artifactory.aql.model.AqlSortTypeEnum;
import org.artifactory.common.StatusEntry;
import org.artifactory.exception.CancelException;
import org.artifactory.fs.FileInfo;
import org.artifactory.md.Properties;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.RepoRepoPath;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.service.flexible.MoveCopyItemInfo;
import org.artifactory.repo.service.flexible.context.MoveCopyContext;
import org.artifactory.repo.service.flexible.listeners.MoveCopyListeners;
import org.artifactory.repo.service.flexible.validators.MoveCopyValidator;
import org.artifactory.sapi.fs.VfsItem;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.storage.fs.MutableVfsFile;
import org.artifactory.storage.fs.MutableVfsItem;
import org.artifactory.storage.fs.lock.SessionLockEntry;
import org.artifactory.storage.fs.session.StorageSessionHolder;
import org.artifactory.util.RepoPathUtils;
import org.jfrog.client.util.PathUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.createWithEmptyResults;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.depth;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.name;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.path;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.property;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.repo;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.type;

/**
 * The class is simple state less helper that help to simplify the move/copy code in the move/copy service
 *
 * @author gidis
 */
public abstract class MoveCopyHelper {

    private static final Logger log = LoggerFactory.getLogger(MoveCopyHelper.class);

    /**
     * returns true if need to stop the move/copy process
     */
    public static boolean isErrorAndFailFast(boolean failFast, MoveMultiStatusHolder finalStatus) {
        return (finalStatus.hasWarnings() || finalStatus.hasErrors()) && failFast;
    }

    /**
     * returns true if need to stop the move/copy process
     */
    public static boolean isError(MoveMultiStatusHolder finalStatus) {
        return (finalStatus.hasWarnings() || finalStatus.hasErrors());
    }

    public static void notifyMoveCopyFiles(
            @NonNull List<MoveCopyItemInfo> elements,
            @NonNull MoveCopyContext context,
            @NonNull MoveMultiStatusHolder status,
            @NonNull List<MoveCopyListeners> moveCopyListeners,
            @NonNull NotifyType notifyType) {

        for (MoveCopyListeners moveCopyListener : moveCopyListeners) {
            if (notifyType == NotifyType.BEFORE) {
                moveCopyListener.notifyBeforeMoveCopyFiles(elements, context, status);
            } else if (notifyType == NotifyType.AFTER) {
                moveCopyListener.notifyAfterMoveCopyFiles(elements, context, status);
            } else {
                throw new IllegalArgumentException(
                        "Unknown notification type " + notifyType + ", supported types are: " +
                                Arrays.toString(NotifyType.values()));
            }

            if (isErrorAndFailFast(context.isFailFast(), status)) {
                break;
            }
        }
    }

    /**
     * The method validates single move/copy of file.
     * It returns:
     * True without errors/warning in staus holder to copy item.
     * True with errors/warning to skip move/copy of an item  and failFast=true to  stop all processes if failFast=true
     * False without errors/warning to skip move/copy of an item if failFast=false and stop all processes if
     * failFast=true
     * False without errors to continue the move/copy
     */
    public static boolean validate(MoveCopyItemInfo element, MoveCopyContext context, MoveMultiStatusHolder status,
            List<MoveCopyValidator> moveCopyValidators) {
        for (MoveCopyValidator moveCopyValidator : moveCopyValidators) {
            if (moveCopyValidator.isInterested(element, context)) {
                boolean valid = moveCopyValidator.validate(element, status, context);
                if (!valid) {
                    return false;
                }
                if (isErrorAndFailFast(context.isFailFast(), status)) {
                    return true;
                }
            }
        }
        return true;
    }

    @NonNull
    public static MoveMultiStatusHolder doExecuteMoveCopyOnBulk(
            @NonNull List<MoveCopyItemInfo> itemsInfos,
            @NonNull MoveCopyContext context,
            @NonNull MultipleArtifactMoveCopier copier,
            @NonNull List<MoveCopyListeners> moveCopyListeners,
            @NonNull List<MoveCopyValidator> validators) {

        MoveMultiStatusHolder status = new MoveMultiStatusHolder();
        List<MoveCopyItemInfo> validItems = new ArrayList<>();
        // validate all items
        for (MoveCopyItemInfo element : itemsInfos) {
            StatusEntry lastError = status.getLastError();
            boolean validItem = validate(element, context, status, validators);
            if (validItem) {
                validItems.add(element);
            }

            if (shouldStopProcess(status, context, lastError)) {
                return status;
            }
        }

        // Notify before move/copy
        notifyMoveCopyFiles(validItems, context, status, moveCopyListeners, NotifyType.BEFORE);
        if (shouldStopProcess(status, context, null)) {
            return status;
        }

        copier.executeOperation(validItems, context, status);
        StorageSessionHolder.getSession().save(); //must save session after copy and before firing interceptors

        if (isErrorAndFailFast(context.isFailFast(), status)) {
            return status;
        }

        // Notify on file successful move. the move of folders in notified at a different time.
        // see org.artifactory.repo.service.flexible.FlatMoveCopyServiceImpl.executeDeleteRootDir
        List<MoveCopyItemInfo> files =
                validItems.stream().filter(item -> item.getSourceItem().isFile()).collect(Collectors.toList());
        notifyMoveCopyFiles(files, context, status, moveCopyListeners, NotifyType.AFTER);

        return status;
    }

    public static boolean shouldStopProcess(MoveMultiStatusHolder status, MoveCopyContext context,
            StatusEntry lastError) {
        @SuppressWarnings("ThrowableResultOfMethodCallIgnored")
        CancelException cancelException = status.getCancelException(lastError);
        return cancelException != null || ((status.hasWarnings() || status.hasErrors()) && context.isFailFast());
    }

    /**
     * Returns true if the repoPath represent trash repo path
     */
    private static boolean isTrash(RepoPath pathToMove) {
        return RepoPathUtils.isTrash(pathToMove);
    }

    /**
     * The method create AQL query that will return all sub folders and files of single root item sorted by dept
     */
    public static AqlApiItem createQuery(RepoPath repoPath, AqlSortTypeEnum order) {
        RepoPath parent = repoPath.getParent();
        AqlBase.AndClause<AqlApiItem> and = AqlApiItem.and();
        AqlApiItem query = createWithEmptyResults().filter(and);
        and.append(AqlApiItem.repo().equal(repoPath.getRepoKey()));
        if (parent == null) {
            // repoPath is Root
            and.append(AqlApiItem.path().matches("*"));
        } else if (parent.isRoot()) {
            // repoPath parent is Root
            AqlBase.OrClause<AqlApiItem> or = AqlApiItem.or();
            AqlBase.AndClause<AqlApiItem> and2 = AqlApiItem.and();
            and2.append(AqlApiItem.name().equal(repoPath.getName()));
            and2.append(AqlApiItem.path().equal("."));
            or.append(and2);
            or.append(AqlApiItem.path().equal(repoPath.getName()));
            or.append(AqlApiItem.path().matches(repoPath.getName() + "/*"));
            and.append(or);
        } else {
            // All other repoPats which are not root and parent is not root
            AqlBase.OrClause<AqlApiItem> or = AqlApiItem.or();
            AqlBase.AndClause<AqlApiItem> and2 = AqlApiItem.and();
            and2.append(AqlApiItem.name().equal(repoPath.getName()));
            and2.append(AqlApiItem.path().equal(parent.getPath()));
            or.append(and2);
            or.append(AqlApiItem.path().equal(repoPath.getPath()));
            or.append(AqlApiItem.path().matches(repoPath.getPath() + "/*"));
            and.append(or);
        }
        and.append(AqlApiItem.type().equal(AqlItemTypeEnum.any.signature));
        query.include(repo(), path(), name(), type(), property().key(), property().value()).addSortElement(depth());
        if (AqlSortTypeEnum.asc == order) {
            query.asc();
        } else {
            query.desc();
        }
        return query;
    }

    @NonNull
    public static List<MoveCopyItemInfo> getMoveCopyItems(
            @NonNull MoveCopyContext context,
            @NonNull Map<RepoPath, Properties> itemsAndProperties,
            @NonNull InternalRepositoryService repositoryService) {

        List<MoveCopyItemInfo> returnList = new ArrayList<>();

        // helper data structures
        Map<RepoPath, RepoPath> targetToSourcePaths = new HashMap<>();
        List<RepoPath> sources = new ArrayList<>();
        List<RepoPath> targets = new ArrayList<>();

        // prepare data
        itemsAndProperties.forEach((repoPath, properties) -> {
            // Calculate the final target repo path
            String contextTargetPath = context.getTargetPath();
            String baseTargetPath = StringUtils.isBlank(contextTargetPath) ? "/" : contextTargetPath;
            // Slash at the means that the user have explicitly ask to restore into dir
            boolean folder = baseTargetPath.endsWith("/");
            // Clean slashes to avoid duplicate slashes
            baseTargetPath = PathUtils.trimSlashes(baseTargetPath);
            String diffPath;
            if (folder) {
                diffPath = getDiffPath(context.getSourceRepoPath(), repoPath, true);
            } else {
                diffPath = getDiffPath(context.getSourceRepoPath(), repoPath, false);
            }
            // Create target path
            String targetPath = baseTargetPath;
            if (StringUtils.isNotEmpty(diffPath)) {
                targetPath += "/" + diffPath;
            }
            if (repoPath.isFolder()) {
                targetPath += "/";
            }

            if (StringUtils.isNotBlank(context.getPathPrefixToIgnore())) {
                targetPath = targetPath.replaceFirst(context.getPathPrefixToIgnore(), "");
            }
            // Make sure to remove duplicate slashes
            targetPath = PathUtils.removeDuplicateSlashes(targetPath);
            String targetRepo =
                    StringUtils.isNotBlank(context.getTargetKey()) ? context.getTargetKey() : repoPath.getRepoKey();

            RepoPath targetRepoPath = InternalRepoPathFactory.create(targetRepo, targetPath);

            sources.add(repoPath);
            targets.add(targetRepoPath);

            targetToSourcePaths.put(targetRepoPath, repoPath);
        });

        // we assume that we have 1 source repo and 1 target repo
        RepoRepoPath<LocalRepo> sourceRrp = repositoryService.getRepoRepoPath(sources.get(0));
        Map<RepoPath, Optional<VfsItem>> sourceItems = sourceRrp.getRepo().getImmutableFsItems(sources);

        RepoRepoPath<LocalRepo> targetRrp = repositoryService.getRepoRepoPath(targets.get(0));
        Map<RepoPath, Optional<VfsItem>> targetItems = targetRrp.getRepo().getImmutableFsItems(targets);

        adjustTargets(sourceItems, targetItems, targetToSourcePaths, targetRrp, context);

        // prepare the MoveCopyItemInfo using the Immutable items
        targetItems.forEach((targetRepoPath, optionalTargetItem) -> {
            VfsItem targetItem = optionalTargetItem.orElse(null);

            RepoPath sourceRepoPath = targetToSourcePaths.get(targetRepoPath);
            VfsItem sourceItem = sourceItems.get(sourceRepoPath).orElseThrow();

            Properties properties = itemsAndProperties.get(sourceRepoPath);

            FileInfo targetOriginalFileInfo = null;
            if (targetRepoPath != null && targetItem != null && targetItem.isFile()) {
                targetOriginalFileInfo = (FileInfo) targetItem.getInfo();
            }
            RepoRepoPath<LocalRepo> sourceRepoRepo = repositoryService.getRepoRepoPath(sourceRepoPath);
            RepoRepoPath<LocalRepo> targetRepoRepo = repositoryService.getRepoRepoPath(targetRepoPath);

            returnList.add(new MoveCopyItemInfo(sourceRepoPath, targetRepoPath, sourceItem, targetItem, sourceRepoRepo,
                    targetRepoRepo, targetOriginalFileInfo, null, properties));
        });
        return returnList;
    }

    @NonNull
    private static void adjustTargets(
            @NonNull Map<RepoPath, Optional<VfsItem>> sourceItems,
            @NonNull Map<RepoPath, Optional<VfsItem>> targetItems,
            @NonNull Map<RepoPath, RepoPath> targetToSourcePaths,
            @NonNull RepoRepoPath<LocalRepo> targetRrp,
            @NonNull MoveCopyContext context) {

        List<RepoPath> originalPaths = new ArrayList<>();
        List<RepoPath> adjustedPaths = new ArrayList<>();

        targetItems.forEach((targetRepoPath, optionalTargetItem) -> {
            VfsItem targetItem = optionalTargetItem.orElse(null);
            RepoPath sourceRepoPath = targetToSourcePaths.get(targetRepoPath);
            VfsItem sourceItem = sourceItems.get(sourceRepoPath).orElseThrow();

            if (targetItem != null && context.isUnixStyleBehavior() && targetItem.isFolder() && sourceItem.isFile()) {
                String adjustedPathString = targetRepoPath.getPath() + "/" + sourceItem.getName();
                RepoPath adjustedPath =
                        InternalRepoPathFactory.create(targetRepoPath.getRepoKey(), adjustedPathString, false);
                originalPaths.add(targetRepoPath);
                adjustedPaths.add(adjustedPath);
            }
        });

        Map<RepoPath, Optional<VfsItem>> newTargets = targetRrp.getRepo().getImmutableFsItems(adjustedPaths);
        for (int i = 0; i < adjustedPaths.size(); i++) {
            RepoPath adjusted = adjustedPaths.get(i);
            RepoPath original = originalPaths.get(i);

            targetItems.remove(original);
            RepoPath correspondingSource = targetToSourcePaths.remove(original);

            targetItems.put(adjusted, newTargets.get(adjusted));
            targetToSourcePaths.put(adjusted, correspondingSource);
        }

    }

    public static void overrideTargetFileIfExist(RepoRepoPath<LocalRepo> targetRrp, LocalRepo targetRepo,
            MoveMultiStatusHolder status, boolean basicStatusHolderErrorLog) {
        MutableVfsItem targetItem = targetRepo.getMutableFsItem(targetRrp.getRepoPath());
        if (targetItem != null) {
            // target repository already contains file or folder with the same name, delete it
            log.debug("File {} already exists in target repository. Overriding.", targetRrp.getRepoPath());
            targetItem.delete(new DeleteContext(targetItem.getRepoPath()).triggeredByMove());
            reloadMutable(targetItem.getRepoPath(), targetRrp);
        } else if (basicStatusHolderErrorLog) {
            status.error(String.format("Failed to create mutable item for path: %s", targetRrp.getRepoPath()), log);
        }
    }

    /**
     * When doing override of a file, MutableVfsFile needs to be 'reset' since we are saving the delete of the target
     * file during a transcation, in order to finally save the new item at the final destination safely later on.
     */
    private static void reloadMutable(RepoPath repoPath, RepoRepoPath<LocalRepo> targetRrp) {
        SessionLockEntry lockEntry = StorageSessionHolder.getSession().getLockEntry(repoPath);
        lockEntry.save();
        lockEntry.setWriteFsItem(null);
        MutableVfsFile newMutable = targetRrp.getRepo().createOrGetFile(repoPath);
        lockEntry.setWriteFsItem(newMutable);
    }

    /**
     * In case of trash can the method removes the repo key from the path
     */
    private static RepoPath cleanRepoKeyFromPath(RepoPath pathToMove) {
        String ancestor = PathUtils.stripFirstPathElement(pathToMove.getPath());
        String repoKey = PathUtils.getFirstPathElement(pathToMove.getPath());
        return RepoPathFactory.create(repoKey, ancestor);
    }

    /**
     * The method resolves the relative sub path of the source in relation to the root folder of the copy
     */
    private static String getDiffPath(RepoPath rootSource, RepoPath currentSource, boolean skipName) {
        if (rootSource == null) {
            return currentSource.getPath();
        }
        if (isTrash(rootSource)) {
            rootSource = cleanRepoKeyFromPath(rootSource);
        }
        if (isTrash(currentSource)) {
            currentSource = cleanRepoKeyFromPath(currentSource);
        }
        String[] rootSourceElements = PathUtils.getPathElements(rootSource.getPath());
        String[] currentSourceElements = PathUtils.getPathElements(currentSource.getPath());
        int maxIndex = Math.min(rootSourceElements.length, currentSourceElements.length);
        int index = 0;
        String commonPath = "";
        int skip = skipName ? 1 : 0;
        while (index < maxIndex - skip && rootSourceElements[index].equals(currentSourceElements[index])) {
            commonPath = StringUtils.isBlank(commonPath) ? rootSourceElements[index] :
                    commonPath + "/" + rootSourceElements[index];
            index++;
        }
        return PathUtils.trimSlashes(currentSource.getPath().replaceFirst(commonPath, "")).toString();
    }

    public enum NotifyType {
        BEFORE,
        AFTER
    }

}

