/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.service.flexible.tools;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.md.Properties;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoRepoPath;
import org.artifactory.repo.service.flexible.MoveCopyItemInfo;
import org.artifactory.repo.service.flexible.context.MoveCopyContext;
import org.artifactory.sapi.fs.VfsFile;
import org.artifactory.sapi.fs.VfsFolder;
import org.artifactory.sapi.fs.VfsItem;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.artifactory.storage.db.fs.model.DbMutableFile;
import org.artifactory.storage.fs.MutableVfsFile;
import org.artifactory.storage.fs.MutableVfsFolder;
import org.artifactory.storage.fs.MutableVfsItem;
import org.artifactory.storage.spring.StorageContextHelper;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * The class execute multiple items move/copy task.
 * Note In  case of move folder the source folder will not be deleted, the folder deletion should be executed
 * only after full copy of the entire folder.
 *
 * @author Eli Skoran
 */
@Slf4j
public class MultipleArtifactMoveCopier {

    /**
     * Execute multiple item move/copy
     */
    void executeOperation(@NonNull List<MoveCopyItemInfo> itemsInfo, @NonNull MoveCopyContext context,
            @NonNull MoveMultiStatusHolder status) {

        Map<MoveCopyItemInfo, MutableVfsItem> mutableVfsItems = doMoveCopy(itemsInfo, status, context);

        // set the mutable target items into the itemsInfo so we can use it down the flow
        for (MoveCopyItemInfo itemInfo : itemsInfo) {
            MutableVfsItem mutableVfsItem = mutableVfsItems.get(itemInfo);
            itemInfo.setMutableTargetItem(mutableVfsItem);
        }
    }

    @NonNull
    private Map<MoveCopyItemInfo, MutableVfsItem> doMoveCopy(
            @NonNull List<MoveCopyItemInfo> moveItems,
            @NonNull MoveMultiStatusHolder status,
            @NonNull MoveCopyContext context) {

        Map<MoveCopyItemInfo, MutableVfsItem> mutableMap = new HashMap<>();
        List<MoveCopyItemInfo> filesToCopy = new ArrayList<>();

        for (MoveCopyItemInfo element : moveItems) {
            RepoRepoPath<LocalRepo> targetRrp = element.getTargetRrp();
            VfsItem sourceItem = element.getSourceItem();
            VfsItem targetItem = element.getTargetItem();
            Properties properties = element.getProperties();
            MutableVfsItem mutableVfsItem;
            if (!context.isDryRun()) {
                if (sourceItem.isFolder()) {
                    mutableVfsItem =
                            shallowCopyDirectory((VfsFolder) sourceItem, properties, targetItem, targetRrp, status);
                    List<MoveCopyItemInfo> foldersToDelete = context.getFoldersToDelete();
                    if (!context.isCopy()) {
                        // moved folder should be deleted
                        foldersToDelete.add(element);
                    }
                    mutableMap.put(element, mutableVfsItem);
                } else {
                    filesToCopy.add(element);
                }

            }
        }

        mutableMap.putAll(copyFiles(filesToCopy, status, context));

        if (!context.isCopy()) {
            deleteFiles(filesToCopy);
        }

        return mutableMap;
    }

    @NonNull
    protected Map<MoveCopyItemInfo, MutableVfsItem> copyFiles(
            @NonNull List<MoveCopyItemInfo> copyElements,
            @NonNull MoveMultiStatusHolder status,
            @NonNull MoveCopyContext context) {

        Map<MoveCopyItemInfo, MutableVfsItem> resultMap = new HashMap<>();

        // we assume that the target repo is the same for all items
        RepoRepoPath<LocalRepo> targetRepo = copyElements.get(0).getTargetRrp();

        List<RepoPath> targetRepoPaths =
                copyElements.stream().map(MoveCopyItemInfo::getTargetRepoPath).collect(Collectors.toList());
        List<VfsItem> sourceItems =
                copyElements.stream().map(MoveCopyItemInfo::getSourceItem).collect(Collectors.toList());
        List<Properties> properties =
                copyElements.stream().map(MoveCopyItemInfo::getProperties).collect(Collectors.toList());

        // get the mutableItems
        Map<RepoPath, MutableVfsItem> targetMutables = targetRepo.getRepo().createOrGetFiles(targetRepoPaths);

        // copy file
        copyVfsFiles(sourceItems, targetRepoPaths, targetMutables, properties, context);

        status.artifactsMoved(sourceItems.size());

        copyElements.forEach(element -> {
                    MutableVfsFile targetMutable = (MutableVfsFile) targetMutables.get(element.getTargetRepoPath());
                    resultMap.put(element, targetMutable);
                }
        );

        return resultMap;

    }

    protected void copyVfsFiles(@NonNull List<VfsItem> sourceFiles,
            @NonNull List<RepoPath> targetRepoPaths,
            @NonNull Map<RepoPath, MutableVfsItem> targetMutables,
            @NonNull List<Properties> properties,
            @NonNull MoveCopyContext context) {

        // copy the info and the properties only (stats and watches are not required)
        List<String> sha1s = sourceFiles.stream()
                .map(sourceFile -> ((VfsFile) sourceFile).getSha1())
                .collect(Collectors.toList());

        List<BinaryElement> binaries = StorageContextHelper.get().beanForType(InternalBinaryService.class)
                .getBinaryRecords(sha1s);

        for (int i = 0; i < targetRepoPaths.size(); i++) {
            RepoPath targetRepoPath = targetRepoPaths.get(i);
            DbMutableFile mutableVfsFile = (DbMutableFile) targetMutables.get(targetRepoPath);
            VfsFile vfsFile = (VfsFile) sourceFiles.get(i);
            BinaryElement binaryElement = binaries.get(i);
            Properties itemProperties = properties.get(i);

            mutableVfsFile.setFileBinaryInfo(binaryElement);
            mutableVfsFile.fillInfo(vfsFile.getInfo());

            Properties newProperties = mergeProperties(itemProperties, context);

            mutableVfsFile.setProperties(newProperties, true);
        }
    }

    private void deleteFiles(@NonNull List<MoveCopyItemInfo> itemsToDelete) {
        List<RepoPath> sourceRepoPaths =
                itemsToDelete.stream().map(MoveCopyItemInfo::getSourceRepoPath).collect(Collectors.toList());

        Map<RepoPath, MutableVfsFile> mutableFiles =
                itemsToDelete.get(0).getSourceRrp().getRepo().getMutableFiles(sourceRepoPaths);

        mutableFiles.forEach((key, value) -> {
            if (value != null) {
                value.delete(new DeleteContext(value.getRepoPath()).triggeredByMove().allowAsyncDelete(true));
            }
        });
    }

    @NonNull
    private MutableVfsFolder shallowCopyDirectory(@NonNull VfsFolder sourceFolder,
            @NonNull Properties properties,
            @Nullable VfsItem targetItem,
            @NonNull RepoRepoPath<LocalRepo> targetRrp,
            @NonNull MoveMultiStatusHolder status) {

        LocalRepo targetRepo = targetRrp.getRepo();
        RepoPath targetRepoPath =
                InternalRepoPathFactory.create(targetRepo.getKey(), targetRrp.getRepoPath().getPath());

        if (targetItem == null) {
            log.debug("Creating target folder {}", targetRepoPath);
        } else {
            log.debug("Target folder {} already exist", targetRepoPath);
        }
        MutableVfsFolder targetFolder = targetRepo.createOrGetFolder(targetRepoPath);
        status.folderMoved();
        // copy relevant metadata from source to target
        log.debug("Copying folder metadata to {}", targetRepoPath);
        targetFolder.fillInfo(sourceFolder.getInfo());
        targetFolder.setProperties(properties, true);
        return targetFolder;
    }

    private Properties mergeProperties(Properties properties, MoveCopyContext context) {
        PropertiesImpl newProperties = new PropertiesImpl();
        for (String key : properties.keySet()) {
            newProperties.putAll(key, properties.get(key));
        }
        if (context.getRemoveProps() != null) {
            context.getRemoveProps().forEach(newProperties::removeAll);
        }
        Properties addProperties = context.getAddProps();
        if (addProperties != null) {
            for (String key : addProperties.keySet()) {
                properties.putAll(key, addProperties.get(key));
            }
        }
        return newProperties;
    }
}
