package org.artifactory.storage.db.build.aql;

import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.AqlApiElement;
import org.artifactory.aql.api.domain.sensitive.AqlApiBuild;
import org.artifactory.aql.result.rows.AqlBuild;
import org.artifactory.checksum.ChecksumType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.util.Objects;
import java.util.stream.Stream;

import static org.artifactory.aql.api.internal.AqlBase.or;

@Repository
public class BuildsAql {

    private static final Logger log = LoggerFactory.getLogger(BuildsAql.class);
    AqlService aqlService;

    @Autowired
    public void setAqlService(AqlService aqlService) {
        this.aqlService = aqlService;
    }

    public Stream<AqlBuild> getDependantBuilds(String sha1, String md5) {
        AqlApiElement<AqlApiBuild> queryChecksumFilter = buildChecksumFilterForDependencies(sha1, md5);
        if (queryChecksumFilter == null) {
            return Stream.empty();
        }

        AqlApiBuild buildDependenciesQuery =
                AqlApiBuild.create()
                        .filter(queryChecksumFilter)
                        .distinct(true)
                        .include(
                                AqlApiBuild.name(),
                                AqlApiBuild.module().dependecy().scope(),
                                AqlApiBuild.number(),
                                AqlApiBuild.started(),
                                AqlApiBuild.module().name(),
                                AqlApiBuild.url(),
                                AqlApiBuild.buildRepo()
                        );
        return aqlService.executeQueryLazy(buildDependenciesQuery)
                .asStream(e -> {
                    if (e != null) {
                        log.error("failed to read build item from stream {}", e.getMessage());
                    }
                })
                .filter(Objects::nonNull);
    }

    private AqlApiElement<AqlApiBuild> buildChecksumFilterForDependencies(String sha1, String md5) {
        boolean validSha1 = ChecksumType.sha1.isValid(sha1);
        boolean validMd5 = ChecksumType.md5.isValid(md5);
        if (validMd5 && validSha1) {
            return or(
                    AqlApiBuild.module().dependecy().sha1().equal(sha1),
                    AqlApiBuild.module().dependecy().md5().equal(md5)
            );
        } else if (validMd5) {
            return AqlApiBuild.module().dependecy().md5().equal(md5);
        } else if (validSha1) {
            return AqlApiBuild.module().dependecy().sha1().equal(sha1);
        } else {
            return null;
        }
    }

    public Stream<AqlBuild> getProducedBuilds(String sha1, String md5) {
        AqlApiElement<AqlApiBuild> queryChecksumFilter = buildChecksumFilterForArtifacts(sha1, md5);
        if (queryChecksumFilter == null) {
            return Stream.empty();
        }

        AqlApiBuild buildDependenciesQuery =
                AqlApiBuild.create()
                        .filter(queryChecksumFilter)
                        .distinct(true)
                        .include(
                                AqlApiBuild.name(),
                                AqlApiBuild.number(),
                                AqlApiBuild.started(),
                                AqlApiBuild.module().name(),
                                AqlApiBuild.url(),
                                AqlApiBuild.buildPromotions().status(),
                                AqlApiBuild.buildRepo()
                        );
        return aqlService.executeQueryLazy(buildDependenciesQuery)
                .asStream(e -> {
                    if (e != null) {
                        log.error("failed to read build item from stream {}", e.getMessage());
                    }
                })
                .filter(Objects::nonNull);
    }

    private AqlApiElement<AqlApiBuild> buildChecksumFilterForArtifacts(String sha1, String md5) {
        boolean validSha1 = ChecksumType.sha1.isValid(sha1);
        boolean validMd5 = ChecksumType.md5.isValid(md5);
        if (validMd5 && validSha1) {
            return or(
                    AqlApiBuild.module().artifact().sha1().equal(sha1),
                    AqlApiBuild.module().artifact().md5().equal(md5)
            );
        } else if (validMd5) {
            return AqlApiBuild.module().artifact().md5().equal(md5);
        } else if (validSha1) {
            return AqlApiBuild.module().artifact().sha1().equal(sha1);
        } else {
            return null;
        }
    }
}
