package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.fetch;

import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.support.SupportAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.rest.common.model.RestModel;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.glassfish.jersey.internal.guava.Lists;
import org.jfrog.common.StreamSupportUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.when;

/**
 * @author Omri Ziv
 */
public abstract class RootFetchStrategyTest {

    protected final static String PROJECT_A = "projectA";
    protected final static String DEFAULT_PROJECT = ProjectsService.DEFAULT_PROJECT;

    @Mock
    AuthorizationService authorizationService;
    @Mock
    RepositoryService repoService;
    @Mock
    AddonsManager addonsManager;
    @Mock
    SupportAddon supportAddon;
    @Mock
    AolService aolService;
    @Mock
    private ArtifactoryContext context;
    @Mock
    ArtifactoryHome artifactoryHome;
    @Mock
    ArtifactorySystemProperties artifactorySystemProperties;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.openMocks(this);
        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getLongProperty(any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    return Long.parseLong(constantValue.getDefValue().trim());
                });
        when(artifactorySystemProperties.getProperty(any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    if (constantValue == ConstantValues.orderTreeBrowserRepositoriesByType) {
                        return "distribution,local,remote,virtual";
                    }
                    return constantValue.getDefValue().trim();
                });
        ArtifactoryContextThreadBinder.bind(context);
        when(context.getRepositoryService())
                .thenReturn(repoService);
        when(context.getAuthorizationService())
                .thenReturn(authorizationService);
        when(context.beanForType(AddonsManager.class))
                .thenReturn(addonsManager);
        when(context.beanForType(RepositoryService.class))
                .thenReturn(repoService);
        when(context.beanForType(AolService.class))
                .thenReturn(aolService);
        when(aolService.isAol())
                .thenReturn(false);
        when(addonsManager.addonByType(SupportAddon.class))
                .thenReturn(supportAddon);
        when(supportAddon.isSupportAddonEnabled())
                .thenReturn(false);
        when(authorizationService.userHasPermissionsOnRepositoryRoot(any()))
                .thenReturn(true);
        when(authorizationService.isAdmin())
                .thenReturn(true);
    }

    @AfterMethod
    public void unbind() {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
    }

    protected void createMocks() {
        List<LocalRepoDescriptor> mavenLocalRepoDescriptor = createLocalRepoDescriptors(RepoType.Maven, 10);
        List<LocalRepoDescriptor> dockerLocalRepoDescriptor = createLocalRepoDescriptors(RepoType.Docker, 5);
        List<LocalRepoDescriptor> npmLocalRepoDescriptor = createLocalRepoDescriptors(RepoType.Npm, 7);
        VirtualRepoDescriptor virtualRepoDescriptor = new VirtualRepoDescriptor();
        virtualRepoDescriptor.setKey("npm-virtual");
        virtualRepoDescriptor.setType(RepoType.Npm);
        RemoteRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setKey("npm-remote");
        remoteRepoDescriptor.setType(RepoType.Npm);
        FederatedRepoDescriptor federatedRepoDescriptor = new FederatedRepoDescriptor();
        federatedRepoDescriptor.setKey("gen-fed");
        federatedRepoDescriptor.setType(RepoType.Generic);
        Map<RepoType, Integer> repoTypeRepoDescriptorMap = Maps.newTreeMap(new RepoType.RepoNameComparator());
        repoTypeRepoDescriptorMap.put(RepoType.Maven, mavenLocalRepoDescriptor.size());
        repoTypeRepoDescriptorMap.put(RepoType.Docker, dockerLocalRepoDescriptor.size());
        repoTypeRepoDescriptorMap.put(RepoType.Npm, npmLocalRepoDescriptor.size());
        List<LocalRepoDescriptor> localRepoDescriptor = Lists.newArrayList(Iterables.concat(mavenLocalRepoDescriptor,
                dockerLocalRepoDescriptor, npmLocalRepoDescriptor));
        List<RepoDescriptor> npmDescriptors = Lists.newArrayList(npmLocalRepoDescriptor);
        npmDescriptors.add(remoteRepoDescriptor);
        npmDescriptors.add(virtualRepoDescriptor);

        Map<Character, List<String>> characterListMap =
                createCharacterListMap(Lists.newArrayList(localRepoDescriptor));

        LocalRepoDescriptor trashDescriptor = new LocalRepoDescriptor();
        trashDescriptor.setKey("trash");
        when(repoService.getFederatedRepoDescriptorsByProject(DEFAULT_PROJECT)).thenReturn(
                Collections.singletonList(federatedRepoDescriptor));
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Maven, DEFAULT_PROJECT))
                .thenReturn(Lists.newArrayList(mavenLocalRepoDescriptor));
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Docker, DEFAULT_PROJECT))
                .thenReturn(Lists.newArrayList(dockerLocalRepoDescriptor));
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Npm, DEFAULT_PROJECT))
                .thenReturn(Lists.newArrayList(npmDescriptors));
        when(repoService.getRepoDescriptorByPackageTypeCountAndProject(DEFAULT_PROJECT))
                .thenReturn(repoTypeRepoDescriptorMap);
        when(repoService.getLocalRepoDescriptorsIncludingInfoReposByProject(DEFAULT_PROJECT))
                .thenReturn(localRepoDescriptor);
        when(repoService.getVirtualRepoDescriptorByProject(DEFAULT_PROJECT))
                .thenReturn(List.of(virtualRepoDescriptor));
        when(repoService.getDistributionRepoDescriptorsByProject(DEFAULT_PROJECT))
                .thenReturn(Collections.emptyList());
        when(repoService.getRemoteRepoDescriptorsByProject(DEFAULT_PROJECT))
                .thenReturn(List.of(remoteRepoDescriptor));
        when(repoService.getAllRepoKeysByFirstCharMapAndProject(DEFAULT_PROJECT))
                .thenReturn(characterListMap);
        when(repoService.repoDescriptorByKey(anyString()))
                .thenAnswer(invocationOnMock -> {
                    String repoKey = invocationOnMock.getArgument(0);
                    return StreamSupportUtils.stream(localRepoDescriptor)
                            .filter(repoDescriptor -> repoDescriptor.getKey().equals(repoKey))
                            .findFirst().orElse(null);
                });
        when(repoService.localOrFederatedRepoDescriptorByKey(TrashService.TRASH_KEY))
                .thenReturn(trashDescriptor);
        when(repoService.hasChildren(any()))
                .thenReturn(false);


        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Maven, PROJECT_A))
                .thenReturn(List.of());
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Docker, PROJECT_A))
                .thenReturn(List.of());
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Npm, PROJECT_A))
                .thenReturn(Lists.newArrayList(npmDescriptors));
        Map<RepoType, Integer> repoTypeRepoDescriptorProjectMap = Maps.newTreeMap(new RepoType.RepoNameComparator());
        repoTypeRepoDescriptorProjectMap.put(RepoType.Npm, npmLocalRepoDescriptor.size());
        when(repoService.getRepoDescriptorByPackageTypeCountAndProject(PROJECT_A))
                .thenReturn(repoTypeRepoDescriptorProjectMap);
        when(repoService.getLocalRepoDescriptorsIncludingInfoReposByProject(PROJECT_A))
                .thenReturn(npmLocalRepoDescriptor);
        when(repoService.getVirtualRepoDescriptorByProject(PROJECT_A))
                .thenReturn(List.of(virtualRepoDescriptor));
        when(repoService.getDistributionRepoDescriptorsByProject(PROJECT_A))
                .thenReturn(List.of());
        when(repoService.getRemoteRepoDescriptorsByProject(PROJECT_A))
                .thenReturn(List.of(remoteRepoDescriptor));
        Map<Character, List<String>> projectCharacterListMap =
                createCharacterListMap(Lists.newArrayList(npmLocalRepoDescriptor));
        when(repoService.getAllRepoKeysByFirstCharMapAndProject(PROJECT_A))
                .thenReturn(projectCharacterListMap);
    }

    protected void createMocksWithLocalAndVirtual() {
        List<LocalRepoDescriptor> mavenLocalRepoDescriptor = createLocalRepoDescriptors(RepoType.Maven, 10);
        List<LocalRepoDescriptor> dockerLocalRepoDescriptor = createLocalRepoDescriptors(RepoType.Docker, 5);
        List<LocalRepoDescriptor> npmLocalRepoDescriptor = createLocalRepoDescriptors(RepoType.Npm, 7);

        List<VirtualRepoDescriptor> mavenVirtualRepoDescriptor = createVirtualRepoDescriptors(RepoType.Maven, 3);
        List<VirtualRepoDescriptor> dockerVirtualRepoDescriptor = createVirtualRepoDescriptors(RepoType.Docker, 4);
        List<VirtualRepoDescriptor> npmVirtualRepoDescriptor = createVirtualRepoDescriptors(RepoType.Npm, 1);

        Map<RepoType, Integer> repoTypeRepoDescriptorMap = Maps.newTreeMap(new RepoType.RepoNameComparator());

        List<RepoDescriptor> mavenRepoDescriptor = Lists.newArrayList(
                Iterables.concat(mavenLocalRepoDescriptor, mavenVirtualRepoDescriptor));
        List<RepoDescriptor> dockerRepoDescriptor = Lists.newArrayList(Iterables.concat(dockerLocalRepoDescriptor,
                dockerVirtualRepoDescriptor));
        List<RepoDescriptor> npmRepoDescriptor =
                Lists.newArrayList(Iterables.concat(npmLocalRepoDescriptor, npmVirtualRepoDescriptor));
        List<LocalRepoDescriptor> localRepoDescriptor = Lists.newArrayList(Iterables.concat(mavenLocalRepoDescriptor,
                dockerLocalRepoDescriptor, npmLocalRepoDescriptor));
        List<VirtualRepoDescriptor> virtualRepoDescriptor =
                Lists.newArrayList(Iterables.concat(mavenVirtualRepoDescriptor,
                        dockerVirtualRepoDescriptor, npmVirtualRepoDescriptor));

        List<RepoDescriptor> allRepoDescriptor =
                Lists.newArrayList(Iterables.concat(localRepoDescriptor, virtualRepoDescriptor));
        Map<Character, List<String>> characterListMap =
                createCharacterListMap(allRepoDescriptor);
        repoTypeRepoDescriptorMap.put(RepoType.Maven, mavenRepoDescriptor.size());
        repoTypeRepoDescriptorMap.put(RepoType.Docker, dockerRepoDescriptor.size());
        repoTypeRepoDescriptorMap.put(RepoType.Npm, npmRepoDescriptor.size());
        LocalRepoDescriptor trashDescriptor = new LocalRepoDescriptor();
        trashDescriptor.setKey("trash");


        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Maven, DEFAULT_PROJECT))
                .thenReturn(mavenRepoDescriptor);
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Docker, DEFAULT_PROJECT))
                .thenReturn(dockerRepoDescriptor);
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Npm, DEFAULT_PROJECT))
                .thenReturn(npmRepoDescriptor);
        when(repoService.getRepoDescriptorByPackageTypeCountAndProject(DEFAULT_PROJECT))
                .thenReturn(repoTypeRepoDescriptorMap);
        when(repoService.getLocalRepoDescriptorsIncludingInfoReposByProject(DEFAULT_PROJECT))
                .thenReturn(localRepoDescriptor);
        when(repoService.getVirtualRepoDescriptorByProject(DEFAULT_PROJECT))
                .thenReturn(virtualRepoDescriptor);
        when(repoService.getDistributionRepoDescriptorsByProject(DEFAULT_PROJECT))
                .thenReturn(Collections.emptyList());
        when(repoService.getRemoteRepoDescriptorsByProject(DEFAULT_PROJECT))
                .thenReturn(Collections.emptyList());
        when(repoService.getAllRepoKeysByFirstCharMapAndProject(DEFAULT_PROJECT))
                .thenReturn(characterListMap);
        when(repoService.repoDescriptorByKey(anyString()))
                .thenAnswer(invocationOnMock -> {
                    String repoKey = invocationOnMock.getArgument(0);
                    return StreamSupportUtils.stream(allRepoDescriptor)
                            .filter(repoDescriptor -> repoDescriptor.getKey().equals(repoKey))
                            .findFirst().orElse(null);
                });
        when(repoService.localOrFederatedRepoDescriptorByKey(TrashService.TRASH_KEY))
                .thenReturn(trashDescriptor);
        when(repoService.hasChildren(any()))
                .thenReturn(false);

        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Maven, PROJECT_A))
                .thenReturn(List.of());
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Docker, PROJECT_A))
                .thenReturn(List.of());
        when(repoService.getRepoDescriptorByPackageTypeAndProject(RepoType.Npm, PROJECT_A))
                .thenReturn(npmRepoDescriptor);
        Map<RepoType, Integer> repoTypeRepoDescriptorProjectMap = Maps.newTreeMap(new RepoType.RepoNameComparator());
        repoTypeRepoDescriptorProjectMap.put(RepoType.Npm, npmRepoDescriptor.size());
        when(repoService.getRepoDescriptorByPackageTypeCountAndProject(PROJECT_A))
                .thenReturn(repoTypeRepoDescriptorProjectMap);
        when(repoService.getLocalRepoDescriptorsIncludingInfoReposByProject(PROJECT_A))
                .thenReturn(npmLocalRepoDescriptor);
        when(repoService.getVirtualRepoDescriptorByProject(PROJECT_A))
                .thenReturn(npmVirtualRepoDescriptor);
        when(repoService.getDistributionRepoDescriptorsByProject(PROJECT_A))
                .thenReturn(Collections.emptyList());
        when(repoService.getRemoteRepoDescriptorsByProject(PROJECT_A))
                .thenReturn(Collections.emptyList());
        Map<Character, List<String>> projectCharacterListMap =
                createCharacterListMap(Lists.newArrayList(npmLocalRepoDescriptor));
        when(repoService.getAllRepoKeysByFirstCharMapAndProject(PROJECT_A))
                .thenReturn(projectCharacterListMap);
    }

    protected boolean testRepositoryNodeBefore(List<? extends RestModel> repositoryNodes, String first, String second) {
        List<String> list = StreamSupportUtils.stream(repositoryNodes)
                .filter(repositoryNode -> {
                    return ((RepositoryNode) repositoryNode).getText().equals(first)
                            || ((RepositoryNode) repositoryNode).getText().equals(second);
                })
                .map(repositoryNode -> ((RepositoryNode) repositoryNode).getText())
                .collect(Collectors.toList());
        return list.size() == 2 && list.get(0).equals(first) && list.get(1).equals(second);
    }

    private List<LocalRepoDescriptor> createLocalRepoDescriptors(RepoType repoType, int number) {
        return IntStream.range(1, number + 1)
                .mapToObj(index -> createLocalRepoDescriptor(repoType.getDisplayName() +
                        "-repo" + String.format("%03d", index), repoType))
                .collect(Collectors.toList());
    }

    private LocalRepoDescriptor createLocalRepoDescriptor(String repoKey, RepoType repoType) {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey(repoKey);
        localRepoDescriptor.setType(repoType);
        return localRepoDescriptor;
    }

    private List<VirtualRepoDescriptor> createVirtualRepoDescriptors(RepoType repoType, int number) {
        return IntStream.range(1, number + 1)
                .mapToObj(index -> createVirtualRepoDescriptor(repoType.getDisplayName() +
                        "-virt-repo" + String.format("%03d", index), repoType))
                .collect(Collectors.toList());

    }

    private VirtualRepoDescriptor createVirtualRepoDescriptor(String repoKey, RepoType repoType) {
        VirtualRepoDescriptor virtualRepoDescriptor = new VirtualRepoDescriptor();
        virtualRepoDescriptor.setKey(repoKey);
        virtualRepoDescriptor.setType(repoType);
        return virtualRepoDescriptor;
    }

    private Map<Character, List<String>> createCharacterListMap(List<RepoDescriptor> repoDescriptors) {

        return StreamSupportUtils.stream(repoDescriptors)
                .map(RepoDescriptor::getKey)
                .sorted(String::compareToIgnoreCase)
                .collect(Collectors.groupingBy(repoKey -> repoKey.toLowerCase().charAt(0)));
    }

    @DataProvider
    public Object[][] provideProjectKeys() {
        return new Object[][] {
                {DEFAULT_PROJECT},
                {PROJECT_A}
        };
    }
}
