package org.artifactory.ui.rest.service.general;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.AddonsWebManager;
import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.addon.CoreAddons;
import org.artifactory.addon.OssAddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.addon.license.ArtifactoryBaseLicenseDetails;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.api.subscriptions.SubscriptionsService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.descriptor.trashcan.TrashcanConfigDescriptor;
import org.artifactory.metadata.service.MetadataEventService;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.security.SimpleUser;
import org.artifactory.storage.binstore.service.BinaryService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.ui.rest.model.general.FooterImpl;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.platform.subscription.SubscriptionType;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.artifactory.addon.OssAddonsManager.FREE_LICENSE_KEY_HASH;
import static org.artifactory.addon.OssAddonsManager.JCR_LICENSE_KEY_HASH;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;

/**
 * @author dudim
 */
@Test
public class GetFooterServiceTest extends ArtifactoryHomeBoundTest {

    @Mock
    private ArtifactoryContext context;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
    }

    public void testGetVersionID() {
        GetFooterService footerService = new GetFooterService(null, null, null);

        Assert.assertEquals(footerService.getVersionID(JCR_LICENSE_KEY_HASH), "JCR");
        Assert.assertEquals(footerService.getVersionID(JCR_LICENSE_KEY_HASH + " Online"), "JCR");
        Assert.assertEquals(footerService.getVersionID(FREE_LICENSE_KEY_HASH), "ConanCE");
    }

    public void testGetPlatformIdJcrAol() {
        GetFooterService footerServiceSpy = spy(new GetFooterService(null, null, null));
        AddonsManager addonsManagerMock = mock(AddonsManager.class);
        doReturn(addonsManagerMock).when(footerServiceSpy).getAddonManager();
        when(addonsManagerMock.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.AOL_JCR);
        Assert.assertEquals(footerServiceSpy.getPlatformId(), "JCR");
    }

    @Test(dataProvider = "archiveIndexProvider")
    public void testIsArchiveIndexEnabled(boolean isArchiveIndexEnabled) {
        FooterImpl iModel = executeAndGetFooter(isArchiveIndexEnabled, SubscriptionType.ENTERPRISE_PLUS);
        Assert.assertEquals(iModel.isArchiveIndexEnabled(), isArchiveIndexEnabled);
    }

    @Test(dataProvider = "typesProvider")
    public void testGetVersionInfoFreeTier(SubscriptionType type) {
        FooterImpl iModel = executeAndGetFooter(false, type);
        if (type != null) {
            Assert.assertEquals(iModel.getSubscriptionTierType(), type.getString());
        } else {
            Assert.assertNull(iModel.getSubscriptionTierType());
        }
    }

    public void testGetVersionInfoJcrOnline() {
        GetFooterService footerServiceSpy = spy(new GetFooterService(null, null, null));
        AddonsManager addonsManagerMock = mock(AddonsManager.class);
        doReturn(addonsManagerMock).when(footerServiceSpy).getAddonManager();
        when(addonsManagerMock.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.AOL_JCR);
        String versionInfo = footerServiceSpy.getVersionInfo();
        Assert.assertEquals(versionInfo, JCR_LICENSE_KEY_HASH + " Online");
    }

    public void testGetVersionInfoJcr() {
        GetFooterService footerServiceSpy = spy(new GetFooterService(null, null, null));
        OssAddonsManager addonsManagerMock = mock(OssAddonsManager.class);
        doReturn(addonsManagerMock).when(footerServiceSpy).getAddonManager();
        when(addonsManagerMock.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.JCR);
        when(addonsManagerMock.getLicenseKeyHash(false)).thenReturn(JCR_LICENSE_KEY_HASH);
        String versionInfo = footerServiceSpy.getVersionInfo();
        Assert.assertEquals(versionInfo, JCR_LICENSE_KEY_HASH);
    }

    public void testGetVersionInfoConan() {
        GetFooterService footerServiceSpy = spy(new GetFooterService(null, null, null));
        OssAddonsManager addonsManagerMock = mock(OssAddonsManager.class);
        doReturn(addonsManagerMock).when(footerServiceSpy).getAddonManager();
        when(addonsManagerMock.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.CONAN);
        when(addonsManagerMock.getLicenseKeyHash(false)).thenReturn(FREE_LICENSE_KEY_HASH);
        String versionInfo = footerServiceSpy.getVersionInfo();
        Assert.assertEquals(versionInfo, FREE_LICENSE_KEY_HASH);
    }

    @Test(dataProvider = "provideLicenseTypes")
    public void testGetVersionInfoForOthers(String licenseType, String expectedVersionInfo) {
        GetFooterService footerServiceSpy = spy(new GetFooterService(null, null, null));
        AddonsManager addonsManagerMock = mock(AddonsManager.class);
        doReturn(addonsManagerMock).when(footerServiceSpy).getAddonManager();
        when(addonsManagerMock.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.PRO);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManagerMock);

        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(false);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        when(addonsManagerMock.getLicenseKeyHash(false)).thenReturn("x");
        ArtifactoryBaseLicenseDetails licDetails = new ArtifactoryBaseLicenseDetails();
        licDetails.setType(licenseType);
        when(addonsManagerMock.getProAndAolLicenseDetails()).thenReturn(licDetails);

        String versionInfo = footerServiceSpy.getVersionInfo();
        Assert.assertEquals(versionInfo, expectedVersionInfo);
    }

    @Test
    public void testGetUserEmail() {
        Authentication authentication = mock(Authentication.class);
        SimpleUser simpleUser = mock(SimpleUser.class);
        when(simpleUser.getEmail()).thenReturn("a@a.com");
        when(authentication.getPrincipal()).thenReturn(simpleUser);
        SecurityContext mock = mock(SecurityContext.class);
        SecurityContextHolder.setContext(mock);
        when(mock.getAuthentication()).thenReturn(authentication);
        GetFooterService footerService =
                new GetFooterService(mock(CentralConfigService.class), mock(AuthorizationService.class),
                        mock(MetadataEventService.class));
        Assert.assertEquals(footerService.getUserEmail(), "a@a.com");
        SecurityContextHolder.clearContext();
    }

    @Test
    public void testGetUserEmailNull() {
        SecurityContext mock = mock(SecurityContext.class);
        SecurityContextHolder.setContext(mock);
        when(mock.getAuthentication()).thenReturn(null);
        GetFooterService footerService =
                new GetFooterService(mock(CentralConfigService.class), mock(AuthorizationService.class),
                        mock(MetadataEventService.class));
        Assert.assertNull(footerService.getUserEmail());
        SecurityContextHolder.clearContext();
    }

    @DataProvider
    public Object[][] provideLicenseTypes() {
        return new Object[][] {
                {"Trial", "Artifactory Trial"},
                {"Edge", "Artifactory Edge"},
                {"Edge Trial", "Artifactory Edge Trial"},
                {"Enterprise Plus", "Artifactory Enterprise Plus"},
                {"Enterprise Plus Trial", "Artifactory Enterprise Plus Trial"},
                {"Commercial", "Artifactory Professional"}
        };
    }

    @DataProvider
    public Object[][] typesProvider() {
        return new Object[][] {
                {SubscriptionType.FREE},
                {SubscriptionType.FREE_XRAY},
                {SubscriptionType.PRO_TEAM},
                {SubscriptionType.ENTERPRISE_TEAM},
                {SubscriptionType.ENTERPRISE_PLUS},
                {null}
        };
    }

    @DataProvider
    public Object[][] archiveIndexProvider() {
        return new Object[][] {
                {true},
                {false},
        };
    }

    private void prepareAddOns(GetFooterService footerServiceSpy) {
        AddonsManager addonsManagerMock = mock(AddonsManager.class);
        when(addonsManagerMock.getLicenseTypeForProduct(any())).thenReturn("OSS");
        CoreAddons coreAddons = mock(CoreAddons.class);
        when(addonsManagerMock.addonByType(CoreAddons.class)).thenReturn(coreAddons);
        HaCommonAddon haCommonAddon = mock(HaCommonAddon.class);
        when(addonsManagerMock.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        when(addonsManagerMock.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.PRO);
        doReturn(addonsManagerMock).when(footerServiceSpy).getAddonManager();
    }

    @NotNull
    private CentralConfigService prepareCentralConfigDescriptor(boolean isArchiveIndexEnabled) {
        CentralConfigService centralConfigService = mock(CentralConfigService.class);
        CentralConfigDescriptor centralConfigDescriptor = mock(CentralConfigDescriptor.class);
        TrashcanConfigDescriptor trashcanConfigDescriptor = mock(TrashcanConfigDescriptor.class);
        when(centralConfigDescriptor.getTrashcanConfig()).thenReturn(trashcanConfigDescriptor);
        SecurityDescriptor securityDescriptor = mock(SecurityDescriptor.class);
        when(centralConfigDescriptor.getSecurity()).thenReturn(securityDescriptor);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        when(centralConfigDescriptor.isArchiveIndexEnabled()).thenReturn(isArchiveIndexEnabled);
        return centralConfigService;
    }

    private void prepareContext(SubscriptionType type) {
        AddonsWebManager addonsWebManager = mock(AddonsWebManager.class);
        BinaryService binaryService = mock(BinaryService.class);
        SubscriptionsService subscriptionsService = mock(SubscriptionsService.class);
        when(subscriptionsService.getSubscriptionType()).thenReturn(type);
        when(context.getArtifactoryHome()).thenReturn(homeStub);
        when(context.beanForType(AddonsWebManager.class)).thenReturn(addonsWebManager);
        when(context.beanForType(BinaryService.class)).thenReturn(binaryService);
        when(context.beanForType(SubscriptionsService.class)).thenReturn(subscriptionsService);
    }

    private FooterImpl executeAndGetFooter(boolean isArchiveIndexEnabled, SubscriptionType subscriptionType) {
        CentralConfigService centralConfigService = prepareCentralConfigDescriptor(isArchiveIndexEnabled);
        MetadataEventService metadataEventService = mock(MetadataEventService.class);
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        UserGroupService userGroupService = mock(UserGroupService.class);
        when(authorizationService.isAnonymous()).thenReturn(false);
        prepareContext(subscriptionType);
        GetFooterService footerServiceSpy =
                spy(new GetFooterService(centralConfigService, authorizationService, metadataEventService));
        prepareAddOns(footerServiceSpy);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(true);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryRestResponse artifactoryRestResponse = new ArtifactoryRestResponse();
        footerServiceSpy.execute(null, artifactoryRestResponse);
        return (FooterImpl) artifactoryRestResponse.getIModel();
    }
}